#include <stdio.h>
#include <stdlib.h>
#include "myalloc.h"
#include "testalloc.h"
#include <pthread.h>
#include <string.h>
#include <assert.h>

// Calculate the sum of the sequence 1, 2, ..., n
size_t sum_from_one(size_t n) {
 return n*(n+1)/2;
}

// Remember also to test ranges of pointers returned!!


void basictest() {
	int *i, *j;
	long k = 0;
	void *p = NULL;
	
	printf("Beginning basic test \n");
	
	// 1. Try to allocate/reallocate 0 bytes
	i = myalloc(0);
	assert(i == NULL); // Return value should be the NULL pointer 
	i = myrealloc(p, 0);
	assert(i == NULL); // Return value should be the NULL pointer 
	printf("---part 1\n");

	
	// 2. Try to allocate more memory than available
	i = myalloc(TEST_MEMORY_SIZE);
	assert(i == NULL); // Return value should be the NULL pointer 
	printf("---part 2\n");
	
	// 3. Try to allocate space for two integers
	i = myalloc(sizeof(int));
	j = myalloc(sizeof(int));

	// Compare memory adresses that i and j point to
	assert(i != j); // the allocator has returned the same adress twice!
	myfree(i); // Release integers
	myfree(j); // Release integers
	printf("---part 3\n");
	
	// 4. Allocate an odd number
	i = myalloc(911);
	assert(i != NULL);
	myfree(i);
	printf("---part 4\n");

	// 5. Allocate a big chunk of memory
	i = myalloc(TEST_MEMORY_SIZE - 1000);
	assert(i != NULL);
	myfree(i);
	printf("---part 5\n");

	//6. Reallocate space 
	i = myalloc(sizeof(int));
	j = myrealloc(i, TEST_MEMORY_SIZE);
	assert(j == NULL);
	myfree(i);
	printf("---part 6\n");
	
	printf("Basic test passed \n");
}

void stresstest() {
	char * allocations [MAX_NUMBER_ALLOCATIONS]; // Maximum number of allocations
	char * test;
	int i = 0, j = 0, k = 0;

	printf("Beginning stress test \n");
	
	// Perform 25 overall cycles
	for (j = 0; j < 25; j++) {		
	  if (j % 5 == 0) printf("Doing cycle %i out of 25\n", j);
	  
		// Perform up to MAX_NUMBER_ALLOCATIONS allocations
		for (i = 0; i < MAX_NUMBER_ALLOCATIONS; i++) {
		 allocations[i] = myalloc(900);                     // Allocate 900 bytes
		 assert(allocations[i] != NULL);                    // Check that NULL is not returned
	     memset(allocations[i], (char) (i % 255), 900);     // Set memory to the value i % 255
		}
		
		// Free the MAX_NUMBER_ALLOCATIONS allocations
		for (i = 0; i < MAX_NUMBER_ALLOCATIONS; i++) {
			// Check memory area...
			test = allocations[i];
			
			for (k = 0; k < 900; k++) {
				if (*test != (char)(i % 255)) {
				 printf("Stress test failed, checksum did not match %d \n", (i % 255));
				 exit(1);
				}
			test++; // Do pointer arithmetic, moving the pointer adress sizeof(char)  up
			}
			myfree(allocations[i]); // Free allocated area
		}
	}
	printf("Stress test passed \n");
}

void *test_thread_worker( void *ptr )
{
    int i, j;

	size_t currentsize = 1;    // Size of current test area in memory measued in bytes
	size_t no_of_numbers;      // Number of test numbers
	size_t chksum = 0;         // A checksum
	
	
	for (i = 1; i < 8000; i++) {
	 chksum         = 0;
	 currentsize    = GETNEXTNUMBER(currentsize);
	 no_of_numbers  = currentsize / sizeof(size_t);
	 size_t * test  = myalloc(currentsize);	 

	 //printf("Size: %d \n", currentsize);
	 //memstat();
	 
	 assert(test != NULL);
	 
	 // Fill memory area with data
	 for (j = 0; j < no_of_numbers; j++) test[j] = j+1;
	 
	 // Calculate checksum of memory
	 for (j = 0; j < no_of_numbers; j++) chksum = chksum + test[j];
	 
	 if (chksum != sum_from_one(no_of_numbers))  {
		printf("Thread test not passed \n");
		printf("Allocated memory: %d\n", currentsize);
		printf("Error checksum is %i but should be %i\n", (int) chksum, (int) sum_from_one(no_of_numbers));
		myfree(test);
	    exit(1);
		}
	
  	  myfree(test);
	 }
	
}
	
void threadtest() {
	pthread_t thread1, thread2, thread3;
    int message1 = 1;
    int message2 = 2;
	int i = 0;
	int iret1, iret2, iret3;

	printf("Beginning thread test \n");

	for (i = 0; i < 5; i++) {
		printf("Cycle %i of 5\n", i+1);
	
	     /* Create independent threads each of which will execute function */
	     iret1 = pthread_create( &thread1, NULL, test_thread_worker, &message1);
	     iret2 = pthread_create( &thread2, NULL, test_thread_worker, &message2);
	     
		 pthread_join(thread1, NULL);
	     pthread_join(thread2, NULL); 
	 }
	 
	 
	 printf("Thread test passed \n");
}


void welcome() {
	printf("Welcome to the test program ver 1.0\n");
	printf("On this architecture, an integer is of size %i bytes\n", sizeof(int));
	printf("On this architecture, an size_t is of size %i bytes\n", sizeof(size_t));
	printf("Good luck...\n");
}


int main()
{
	char testmemoryspace[TEST_MEMORY_SIZE];
	
	memset(&testmemoryspace, (char) 0, TEST_MEMORY_SIZE);
	
	welcome();
	
	init_myallocator(&testmemoryspace, TEST_MEMORY_SIZE);

	basictest();
	stresstest();
	threadtest();
	
	printf("Congratulations, all tests passed \n");

	return 0;
}
